/* Animation von Backtracking beim n Damen Problem */
/* benoetigt Java 1.1.7B oder hoeher               */
/*                                                 */
/* (c) Thomas Duebendorfer, Juli 2000              */
/* Fuer Lehrzwecke frei verwendbar und anpassbar.  */
/* Kommerzielle Verwendung verboten.               */
/*                                                 */

import java.awt.*;
import java.awt.event.*;
import java.util.Date;

public class Dame extends Frame implements ActionListener, ItemListener {

MenuItem newDame, solve, quit, about;
Menu AnimationMenu, SchachbrettMenu, DelayMenu, HelpMenu;
CheckboxMenuItem[] size, delay;
DameLsg loesung;
DameCanvas dameCanvas;
int lastX, lastY;
boolean inAnApplet = true;

// Schachbrettgroesse nxn
final static int MINSIZE = 3, MAXSIZE = 15; // Minimum und Maximum
int n=4; // default

// Animationsverzgerung
final static int MINDELAY = 0, DELTADELAY = 50, MAXDELAY = 800; // Minimum und Maximum
int cur_delay = 400;

/*** Code aus Vortrag zu Backtracking (einzig erweitert um Visualisierung) ***/
// Konstante: dame[Spalte] = Zeile = UNDEF, wenn die Spalte keine Dame hat
final int UNDEF = -1;

boolean FindeLoesung(int spalte, DameLsg loesung) {
	// spalte      = aktuelle Schrittzahl und zugleich aktuelle Spalte
	// loesung     = Referenz auf bisherige Teil-Lsung
	int zeile = -1;

	// while(es gibt es noch neue Teil-Lsungsschritte)
	while (zeile != n-1) {
		// Whle einen neuen Teil-Lsungsschritt schritt;
		zeile++;	// Zeilen von oben nach unten der Reihe nach ausprobieren
		
		// Visualisierung
		visualisiereFindeLoesung(spalte,loesung, spalte, zeile, 4);	
		
 		// if (schritt ist gltig), d.h. Dame (zeile, spalte) wird nicht bedroht
		if (loesung.z[zeile] && loesung.d1[zeile+spalte] && loesung.d2[zeile-spalte+n-1]) {
			// Erweitere loesung um schritt
			// Platziere Dame auf (zeile, spalte)
			loesung.dame[spalte] 				= zeile;
			loesung.z[zeile] 					= false;
			loesung.d1[zeile + spalte] 			= false;
			loesung.d2[zeile - spalte  + n - 1] = false;
			
			// Visualisierung
			visualisiereFindeLoesung(spalte,loesung, spalte, zeile, 1);			 

			// if (loesung noch nicht vollstndig) 
			if (spalte != n-1) {
				// rekursiver Aufruf von FindeLoesung
				if (FindeLoesung(spalte+1, loesung)) {
					// Lsung gefunden	
					return true;
				} else {
					// Wir sind in einer Sackgasse:
					// Mache schritt rckgngig: Backtracking
					loesung.dame[spalte] 				= UNDEF;
					loesung.z[zeile] 					= true;
					loesung.d1[zeile + spalte] 			= true;
					loesung.d2[zeile - spalte  + n - 1] = true;
					
					// Visualisierung
					visualisiereFindeLoesung(spalte,loesung, spalte, zeile, 2);
				}
			} else return true; // Lsung gefunden -> fertig		
		}	
	}
	return false; // keine Lsung gefunden
}
/*** Ende des Codes vom Vortrag ***/

/* Zusatz Code */
public Dame() {
// Konstruktor
	newSchachbrett(n); // Datenstrukturen initialisieren
	
	setLayout(new BorderLayout());
	add(dameCanvas = new DameCanvas(this), BorderLayout.CENTER);
	
	// MenuBar mit Menu erzeugen
	MenuBar mb		= new MenuBar();
	AnimationMenu	= new Menu("Animation");
	SchachbrettMenu	= new Menu("Schachbrett");
	DelayMenu		= new Menu("Animationsverzgerung"); 
	HelpMenu		= new Menu("Hilfe");
	
	// MenuItems zum Menu erzeugen
	newDame	     = new MenuItem("Neues Schachbrett");
	solve        = new MenuItem("Aufstellung finden mit Animation");
	quit         = new MenuItem("Beenden");
	about		 = new MenuItem("About ...");

	// Checkbox Menus fuer Schachbrettgroesse hinzufuegen
	size = new CheckboxMenuItem[MAXSIZE];
	for (int i = MINSIZE; i < MAXSIZE; i++) {
		size[i-MINSIZE] = new CheckboxMenuItem(i + "x" + i,false);
		size[i-MINSIZE].addItemListener(this);
		SchachbrettMenu.add(size[i-MINSIZE]);
	}
	size[n-MINSIZE].setState(true);
	
	// Checkbox Menus fuer Animationsverzoegerung hinzufuegen
	delay = new CheckboxMenuItem[(MAXDELAY - MINDELAY) / DELTADELAY + 1];
	for (int i = MINDELAY; i <= MAXDELAY; i = i + DELTADELAY) {
		delay[(i - MINDELAY) / DELTADELAY] = new CheckboxMenuItem(i + " ms",false);
		delay[(i - MINDELAY) / DELTADELAY].addItemListener(this);
		DelayMenu.add(delay[(i - MINDELAY) / DELTADELAY]);
	}
	delay[(cur_delay - MINDELAY) / DELTADELAY].setState(true);
	
	// Ereignisbehandlung hinzufuegen zu MenuItems
	newDame.addActionListener(this);
	solve.addActionListener(this);
	quit.addActionListener(this);
	about.addActionListener(this);
	
	// MenuItems zum Menu hinzufuegen
	AnimationMenu.add(newDame);
	AnimationMenu.add(solve);
	AnimationMenu.addSeparator();
	AnimationMenu.add(quit);
	HelpMenu.add(about);
	
	// Menu zur Menuleiste hinzufuegen	
	mb.add(AnimationMenu);
	mb.add(SchachbrettMenu);
	mb.add(DelayMenu);
	mb.add(HelpMenu);
	setMenuBar(mb);
	
	// Auf Schliessen des Fentsers reagieren
	addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                if (inAnApplet) { 
					dispose(); 
				} else {
                	System.exit(0);
				}
            }
     });

}

public void visualisiereFindeLoesung(int index, DameLsg loesung, int x, int y, int mode) {
// Visualisierung von FindeLoesung
	Graphics g = dameCanvas.getGraphics();
	
	int w = dameCanvas.getSize().width;
    int h = dameCanvas.getSize().height;
	
	int xoff = 5, yoff = 5;
	int xsize = (w - 2*xoff) / n;
	int ysize = (h - 2*yoff) / (n+DameCanvas.legendSize);
	xoff = (w - n * xsize) / 2;
	yoff = (h - (n+DameCanvas.legendSize) * ysize) / 2;
	g.translate(xoff,yoff);
	
	// Loesche Rahmen des letzten aktuellen Feldes
	if((lastX != -1) && (lastY != -1)) {
		// Loesche Rahmen um aktuelles Feld nach einer Pause
		if ((lastX+lastY) % 2 == 1) {
			g.setColor(Color.green);
		} else {
			g.setColor(Color.white);
		}
		g.drawRect(lastX*xsize + 5, lastY*ysize + 5, xsize-9, ysize-9);
		g.drawRect(lastX*xsize + 4, lastY*ysize + 4, xsize-7, ysize-7);	
	};
	
	// Zeichne Rahmen um aktuelles Feld
	g.setColor(Color.black);
	g.drawRect(x*xsize + 5, y*ysize + 5, xsize-9, ysize-9);
	g.drawRect(x*xsize + 4, y*ysize + 4, xsize-7, ysize-7);	
	lastX = x;
	lastY = y;
	
	if (mode == 1) {
		// Hellgraues Rechteck zeichnen		
		g.setColor(Color.lightGray); 
		g.fillRect(x*xsize + 6, y*ysize + 6, xsize-10, ysize-10);	

		// Beschrifte mit Schrittzahl
		mode = 3;
	} 
	
	if (mode == 2) {
		// Entferne Dame
		if ((x+y) % 2 == 1) {
			g.setColor(Color.green);
		} else {
			g.setColor(Color.white);
		}
		g.fillRect(x*xsize + 1, y*ysize + 1, xsize-1, ysize-1);	
	}
	
	if (mode == 3) {
		// Beschrifte mit Schrittzahl
		g.setColor(Color.black);
		g.setFont(new Font("Helvetica", Font.PLAIN, ysize/2));
		String str = "" + index;
		int strWidth = g.getFontMetrics().stringWidth(str);
		int xLoc = x*xsize+xsize/2-strWidth/2;
		g.drawString(str, xLoc, y*ysize+ysize/2+ysize/2/2);	
	}
	
	// Pause 
	sleep(cur_delay);
}


public void sleep(long milliseconds) {
// Wartet die angegebene Anzahl Millisekunden
	Date date;
	long now, start;
	
	date = new Date();
	start = date.getTime();
	now = start;
	while (start + milliseconds > now) {
		/* busy waiting */
		date = new Date();
		now = date.getTime();
	}
}

public void actionPerformed(ActionEvent e) {
// Menu-Ereignisse behandeln
	if (e.getSource() == about) {
		String text = "Backtracking beim n Damen Problem\n";
		text += "(c) Juli 2000, Thomas Duebendorfer, thomas@duebendorfer.ch\n";
		text += "Programm und Quellcode zu Lehrzwecken frei verwendbar.\n";
		text += "Kommerzielle Verwendung nur mit schriftlicher Einwilligung.\n\n";
		text += "Algorithmus:\n";
		text += "Es wird versucht, die n Damen auf einem nxn Schachbrett nacheinander ";
		text += "in die Spalten zu platzieren, ohne dass sich diese bedrohen. Konnte eine";
		text += "Dame nicht in der aktuellen Spalte platziert werden, so werden die ";
		text += "Schritte mit Backtracking zurckgenommen.";
		
		SimpleDialog dlg = new SimpleDialog(this, "About", true, text, "OK");
		dlg.show();
	}
	if (e.getSource() == newDame) {
		newSchachbrett(n);
		dameCanvas.repaint();
	}
	if (e.getSource() == solve) {
		newSchachbrett(n);
		dameCanvas.repaint();	
		solve();
	}
	if (e.getSource() == quit) { System.exit(0); }
}

public void itemStateChanged(ItemEvent e) {
	// Testen, ob Brettgroesse durch Menuauswahl veraendert wurde
	for (int i = MINSIZE; i < MAXSIZE; i++) {
		if (e.getSource() == size[i-MINSIZE]) {
			size[n-MINSIZE].setState(false);
			size[i-MINSIZE].setState(true);
			newSchachbrett(i);
			dameCanvas.repaint(); 
			break; // end for loop
		}
	}

	// Testen, ob Animationsverzoegerung durch Menuauswahl veraendert wurde	
	for (int i = MINDELAY; i <= MAXDELAY; i = i + DELTADELAY) {
		if (e.getSource() == delay[(i - MINDELAY) / DELTADELAY] ) {
			delay[(cur_delay - MINDELAY) / DELTADELAY].setState(false);
			delay[(i - MINDELAY) / DELTADELAY].setState(true);
			cur_delay = i;
			break;
		}
	}
}

void solve() {
// n Damen Problem mit Backtracking loesen und dabei animieren
	setTitle("Backtracking beim n Damen Problem - Animation ...");

	// Disable Menus during animation
	AnimationMenu.setEnabled(false);
	SchachbrettMenu.setEnabled(false);
	DelayMenu.setEnabled(false);
	HelpMenu.setEnabled(false);
	
	dameCanvas.repaint();
	// Startfeld markieren
	// Visualisierung
	visualisiereFindeLoesung(0,loesung, 0, 0, 1);
	if (FindeLoesung(0, loesung)) {
		// Die gefundene Lsung steht in loesung.dame als
		// Position der Dame i von spalte i in zeile loesung.dame[i]
		System.out.println("Aufstellung gefunden!");
		setTitle("Backtracking beim n Damen Problem - Aufstellung gefunden");
		System.out.println("Die Damen stehen auf (spalte, zeile):");
		for(int i = 0; i < n; i++) {
				String s1 = new String("    " + i + ". Dame: ");
				String s2 = new String("    " + i);
				String s3 = new String("    " + loesung.dame[i]);
				String s4 = s1.substring(s1.length()-11) + "(" + s2.substring(s2.length()-4);
				s4 += "," + s3.substring(s3.length()-4) + ")";
				System.out.println(s4);
		}
	} else {
		// Keine Lsung gefunden!
		System.out.println("Keine Aufstellung gefunden!");
		setTitle("Backtracking beim n Damen Problem - Keine Aufstellung gefunden");
	}
	
	repaint();
	
	// Reenable Menus during animation
	AnimationMenu.setEnabled(true);
	SchachbrettMenu.setEnabled(true);
	DelayMenu.setEnabled(true);
	HelpMenu.setEnabled(true);
}

void newSchachbrett(int size) {
// Neues Labyrinth erzeugen und Variable "loesung" zuruecksetzen
	n = size;
	loesung   = new DameLsg(n); // Datenstruktur fuer Loesung erzeugen
		
	for(int i=0; i<n; i++) {
		for(int j=0; j<n; j++) {
			loesung.d1[i+j] = true;
			loesung.d2[i-j+n-1] = true;
		}
		loesung.z[i] = true;
		loesung.dame[i] = UNDEF;
	}
	setTitle("Backtracking beim n Damen Problem");
	lastX = -1;
	lastY = -1;
}
	
public static void main(String[] args) {
    Dame window = new Dame();
	window.inAnApplet = false;

    window.setTitle("Backtracking beim n Damen Problem");
    window.pack(); 
	window.setVisible(true);
}


}